#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2019 Oliver Maier

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import numpy as np
import time
import os
import h5py

from tkinter import filedialog
from tkinter import Tk
import matplotlib.pyplot as plt
import importlib

import pyopencl as cl
import pyopencl.array as clarray


import argparse

from pyqmri._helper_fun import _goldcomp as goldcomp
from pyqmri._helper_fun._est_coils import est_coils
from pyqmri._helper_fun import _utils as utils

DTYPE = np.complex64
DTYPE_real = np.float32


def start_recon(args):
    sig_model_path = os.path.normpath(args.sig_model)
    if len(sig_model_path.split(os.sep)) > 1:
        spec = importlib.util.spec_from_file_location(
            sig_model_path.split(os.sep)[-1], sig_model_path)
        sig_model = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(sig_model)
    else:
        sig_model = importlib.import_module(
            "pyqmri._models."+str(sig_model_path))
    if int(args.streamed) == 1:
        import pyqmri._irgn._reco_streamed as optimizer
    else:
        import pyqmri._irgn._reco as optimizer
    np.seterr(divide='ignore', invalid='ignore')

# Create par struct to store everyting
    par = {}
###############################################################################
# Select input file ###########################################################
###############################################################################
    if args.file == '':
        select_file = True
        while select_file is True:
            root = Tk()
            root.withdraw()
            root.update()
            file = filedialog.askopenfilename()
            root.destroy()
            if not file == () and \
               not file.endswith((('.h5'), ('.hdf5'))):
                print("Please specify a h5 file. Press cancel to exit.")
            elif file == ():
                print("Exiting...")
                return 0
            else:
                select_file = False
    else:
        if not args.file.endswith((('.h5'), ('.hdf5'))):
            print("Please specify a h5 file. ")
            return 0
        file = args.file

    name = os.path.normpath(file)
    fname = name.split(os.sep)[-1]

    par["file"] = h5py.File(file)
#    del par["file"]["Coils"]
###############################################################################
# Read Data ###################################################################
###############################################################################
    reco_Slices = args.slices
    dimX, dimY, NSlice = ((par["file"].attrs['image_dimensions']).astype(int))
    if reco_Slices == -1:
        reco_Slices = NSlice
    off = 0

    data = (par["file"]['real_dat'][
      ..., int(NSlice/2)-int(np.floor((reco_Slices)/2))+off:
      int(NSlice/2)+int(np.ceil(reco_Slices/2))+off, :, :].astype(DTYPE)
            + 1j*par["file"]['imag_dat'][
               ..., int(NSlice/2)-int(np.floor((reco_Slices)/2))+off:
               int(NSlice/2)+int(np.ceil(reco_Slices/2))+off,
               :, :].astype(DTYPE))

    dimreduction = 0
    if args.trafo:
        par["traj"] = par["file"]['real_traj'][()].astype(DTYPE) + \
                      1j*par["file"]['imag_traj'][()].astype(DTYPE)

        par["dcf"] = np.sqrt(np.array(goldcomp.cmp(
                         par["traj"]), dtype=DTYPE_real)).astype(DTYPE_real)
        par["dcf"] = np.require(np.abs(par["dcf"]),
                                DTYPE_real, requirements='C')
    else:
        par["traj"] = None
        par["dcf"] = None
    if np.max(utils.prime_factors(data.shape[-1])) > 13:
        print("Samples along the spoke need to have their largest prime factor"
              " to be 13 or lower. Finding next smaller grid.")
        dimreduction = 2
        while np.max(utils.prime_factors(data.shape[-1]-dimreduction)) > 13:
            dimreduction += 2
        print('Decrease grid size by %i' % dimreduction)
        data = np.require(data[..., int(dimreduction/2):
                               data.shape[-1]-int(dimreduction/2)],
                          requirements='C')
        dimX -= dimreduction
        dimY -= dimreduction
        par["traj"] = np.require(par["traj"][
            ..., int(dimreduction/2):
            par["traj"].shape[-1]-int(dimreduction/2)], requirements='C')
        par["dcf"] = np.sqrt(np.array(goldcomp.cmp(par["traj"]),
                                      dtype=DTYPE_real)).astype(DTYPE_real)
        par["dcf"] = np.require(np.abs(par["dcf"]),
                                DTYPE_real, requirements='C')
###############################################################################
# FA correction ###############################################################
###############################################################################
    if "fa_corr" in list(par["file"].keys()):
        print("Using provied flip angle correction.")
        par["fa_corr"] = np.flip(par["file"]['fa_corr'][()].astype(DTYPE), 0)[
            int(NSlice/2)-int(np.floor((reco_Slices)/2)):
            int(NSlice/2)+int(np.ceil(reco_Slices/2)),
            ...]

        par["fa_corr"] = par["fa_corr"][
            :,
            int(dimreduction/2):par["fa_corr"].shape[-2]-int(dimreduction/2),
            int(dimreduction/2):par["fa_corr"].shape[-1]-int(dimreduction/2)]

        # Recheck if shifted/transposed correctly
        par["fa_corr"] = np.require((np.transpose(par["fa_corr"], (0, 2, 1))),
                                    requirements='C')
    elif "interpol_fa" in list(par["file"].keys()):
        print("Using provied flip angle correction.")
        par["fa_corr"] = np.flip(par["file"]['interpol_fa'][()].astype(DTYPE),
                                 0)[
            int(NSlice/2)-int(np.floor((reco_Slices)/2)):
            int(NSlice/2)+int(np.ceil(reco_Slices/2)),
            int(dimreduction/2):-int(dimreduction/2),
            int(dimreduction/2):-int(dimreduction/2)]

        # Recheck if shifted/transposed correctly
        par["fa_corr"] = np.require((np.transpose(par["fa_corr"], (0, 2, 1))),
                                    requirements='C')
    else:
        print("No flip angle correction provided/used.")
    if data.ndim == 5:
        [NScan, NC, reco_Slices, Nproj, N] = data.shape
    elif data.ndim == 4 and "IRLL" in args.sig_model:
        print("4D Data passed and IRLL model used. Reordering Projections "
              "into 8 Spokes/Frame")
        [NC, reco_Slices, Nproj, N] = data.shape
        Nproj_new = 8
        NScan = np.floor_divide(Nproj, Nproj_new)
        par["Nproj_measured"] = Nproj
        Nproj = Nproj_new
        data = np.require(np.transpose(np.reshape(data[..., :Nproj*NScan, :],
                                       (NC, reco_Slices, NScan, Nproj, N)),
                                       (2, 0, 1, 3, 4)), requirements='C')
        par["traj"] = np.require(np.reshape(par["traj"][:Nproj*NScan, :],
                                 (NScan, Nproj, N)), requirements='C')
        par["dcf"] = np.sqrt(np.array(goldcomp.cmp(par["traj"]),
                                      dtype=DTYPE_real)).astype(DTYPE_real)
        par["dcf"] = np.require(np.abs(par["dcf"]), DTYPE_real,
                                requirements='C')
    else:
        print("Wrong data dimension / model inkompatible. Returning")
        return
###############################################################################
# Set sequence related parameters #############################################
###############################################################################
    for att in par["file"].attrs:
        par[att] = par["file"].attrs[att]

    par["NC"] = NC
    par["dimY"] = dimY
    par["dimX"] = dimX
    par["NSlice"] = reco_Slices
    par["NScan"] = NScan
    par["N"] = N
    par["Nproj"] = Nproj
    if args.streamed:
        par["par_slices"] = args.par_slices

    par["unknowns_TGV"] = sig_model.unknowns_TGV
    par["unknowns_H1"] = sig_model.unknowns_H1
    par["unknowns"] = par["unknowns_TGV"]+par["unknowns_H1"]

    if not args.trafo:
        tmp = np.ones_like(np.abs(data))
        tmp[np.abs(data) == 0] = 0
        par['mask'] = np.reshape(tmp, (data.shape)).astype(DTYPE_real)
        del tmp
    else:
        par['mask'] = None
###############################################################################
# Create OpenCL Context and Queues ############################################
###############################################################################
    platforms = cl.get_platforms()
    par["GPU"] = False
    par["Platform_Indx"] = 0
    if args.use_GPU:
        for j in range(len(platforms)):
            if platforms[j].get_devices(device_type=cl.device_type.GPU):
                print("GPU OpenCL platform <%s> found "
                      "with %i device(s) and OpenCL-version <%s>"
                      % (str(platforms[j].get_info(cl.platform_info.NAME)),
                         len(platforms[j].get_devices(
                            device_type=cl.device_type.GPU)),
                         str(platforms[j].get_info(cl.platform_info.VERSION))))
                par["GPU"] = True
                par["Platform_Indx"] = j
    if not par["GPU"]:
        if args.use_GPU:
            print("No GPU OpenCL platform found. Falling back to CPU.")
        for j in range(len(platforms)):
            if platforms[j].get_devices(device_type=cl.device_type.CPU):
                print("CPU OpenCL platform <%s> found "
                      "with %i device(s) and OpenCL-version <%s>"
                      % (str(platforms[j].get_info(cl.platform_info.NAME)),
                         len(platforms[j].get_devices(
                            device_type=cl.device_type.GPU)),
                         str(platforms[j].get_info(cl.platform_info.VERSION))))
                par["GPU"] = False
                par["Platform_Indx"] = j

    par["ctx"] = []
    par["queue"] = []
    if type(args.devices) == int:
        args.devices = [args.devices]
    if args.streamed:
        if len(args.devices) == 1 and args.devices[0] == -1:
            num_dev = []
            for j in range(len(platforms[par["Platform_Indx"]].get_devices())):
                num_dev.append(j)
            par["num_dev"] = num_dev
        else:
            num_dev = args.devices
            par["num_dev"] = num_dev
    else:
        num_dev = args.devices
        par["num_dev"] = num_dev
    for device in num_dev:
        dev = []
        dev.append(platforms[par["Platform_Indx"]].get_devices()[device])
        tmp = cl.Context(dev)
        par["ctx"].append(tmp)
        par["queue"].append(
         cl.CommandQueue(
          tmp,
          platforms[par["Platform_Indx"]].get_devices()[device],
          properties=cl.command_queue_properties.OUT_OF_ORDER_EXEC_MODE_ENABLE
          | cl.command_queue_properties.PROFILING_ENABLE))
        par["queue"].append(
         cl.CommandQueue(
          tmp,
          platforms[par["Platform_Indx"]].get_devices()[device],
          properties=cl.command_queue_properties.OUT_OF_ORDER_EXEC_MODE_ENABLE
          | cl.command_queue_properties.PROFILING_ENABLE))
        par["queue"].append(
         cl.CommandQueue(
          tmp,
          platforms[par["Platform_Indx"]].get_devices()[device],
          properties=cl.command_queue_properties.OUT_OF_ORDER_EXEC_MODE_ENABLE
          | cl.command_queue_properties.PROFILING_ENABLE))
        par["queue"].append(
         cl.CommandQueue(
          tmp,
          platforms[par["Platform_Indx"]].get_devices()[device],
          properties=cl.command_queue_properties.OUT_OF_ORDER_EXEC_MODE_ENABLE
          | cl.command_queue_properties.PROFILING_ENABLE))
###############################################################################
# Coil Sensitivity Estimation #################################################
###############################################################################
    est_coils(data, par, par["file"], args)
###############################################################################
# Standardize data ############################################################
###############################################################################
    [NScan, NC, NSlice, Nproj, N] = data.shape
    if args.trafo:
        if par["file"].attrs['data_normalized_with_dcf']:
            pass
        else:
            data = data*(par["dcf"])
###############################################################################
# generate nFFT  ##############################################################
###############################################################################
    if NC == 1:
        par['C'] = np.ones((data[0, ...].shape), dtype=DTYPE)
    else:
        par['C'] = par['C'].astype(DTYPE)

    FFT = utils.NUFFT(par, trafo=args.trafo)

    def nFTH(x, fft, par):
        siz = np.shape(x)
        result = np.zeros((par["NC"], par["NSlice"], par["NScan"],
                           par["dimY"], par["dimX"]), dtype=DTYPE)
        tmp_result = clarray.empty(fft.queue, (par["NScan"], 1, 1,
                                   par["dimY"], par["dimX"]), dtype=DTYPE)
        for j in range(siz[1]):
            for k in range(siz[2]):
                inp = clarray.to_device(fft.queue,
                                        np.require(x[:, j, k, ...]
                                                    [:, None, None, ...],
                                                   requirements='C'))
                fft.adj_NUFFT(tmp_result, inp)
                result[j, k, ...] = np.squeeze(tmp_result.get())
        return np.transpose(result, (2, 0, 1, 3, 4))

    images = np.require(np.sum(nFTH(data, FFT, par) *
                               (np.conj(par["C"])), axis=1),
                        requirements='C')
    del FFT, nFTH

###############################################################################
# Scale data norm  ############################################################
###############################################################################
    if args.imagespace:
        dscale = DTYPE_real(np.sqrt(2*1e3*NSlice) /
                            (np.linalg.norm(images.flatten())))
        par["dscale"] = dscale
        images = images*dscale
    else:
        dscale = (DTYPE_real(np.sqrt(2*1e3*NSlice)) /
                  (np.linalg.norm(data.flatten())))
        par["dscale"] = dscale
        data = data*dscale

    if args.trafo:
        center = int(N*0.1)
        sig = []
        noise = []
        ind = np.zeros((N), dtype=bool)
        ind[int(par["N"]/2-center):int(par["N"]/2+center)] = 1
        for j in range(Nproj):
            sig.append(np.sum(data[..., int(NSlice/2), j, ind] *
                              np.conj(data[..., int(NSlice/2), j, ind])))
            noise.append(np.sum(data[..., int(NSlice/2), j, ~ind] *
                                np.conj(data[..., int(NSlice/2), j, ~ind])))
        sig = (np.sum(np.array(sig)))/np.sum(ind)
        noise = (np.sum(np.array(noise)))/np.sum(~ind)
        SNR_est = np.abs(sig/noise)
        par["SNR_est"] = SNR_est
        print("Estimated SNR from kspace", SNR_est)
    else:
        center = int(N*0.1)
        ind = np.zeros((dimY, dimX), dtype=bool)
        ind[int(par["N"]/2-center):int(par["N"]/2+center),
            int(par["N"]/2-center):int(par["N"]/2+center)] = 1
        ind = np.fft.fftshift(ind)
        sig = np.sum(data[..., int(NSlice/2), ind] *
                     np.conj(data[..., int(NSlice/2), ind]))/np.sum(ind)
        noise = np.sum(data[..., int(NSlice/2), ~ind] *
                       np.conj(data[..., int(NSlice/2), ~ind]))/np.sum(~ind)
        SNR_est = np.abs(sig/noise)
        par["SNR_est"] = SNR_est
        print("Estimated SNR from kspace", SNR_est)

    opt = optimizer.ModelReco(par, args.trafo,
                              imagespace=args.imagespace)

    if args.imagespace:
        opt.data = images
    else:
        opt.data = data
###############################################################################
# ratio of z direction to x,y, important for finite differences ###############
###############################################################################
    opt.dz = args.dz
###############################################################################
# Start Reco ##################################################################
###############################################################################
    if args.type == '3D':
        #######################################################################
        # Init forward model and initial guess ################################
        #######################################################################
        model = sig_model.Model(par, images)
        par["file"].close()
        #######################################################################
        # IRGN - TGV Reco #####################################################
        #######################################################################
        if "TGV" in args.reg or args.reg == 'all':
            result_tgv = []
            opt.model = model
            ###################################################################
            # IRGN Params #####################################################
            ###################################################################
            opt.irgn_par = utils.read_config(args.config, "3D_TGV")
            opt.execute(TV=0, imagespace=args.imagespace)
            result_tgv.append(opt.result)
            plt.close('all')
            res_tgv = opt.gn_res
            res_tgv = np.array(res_tgv)/(opt.irgn_par["lambd"]*NSlice)
        #######################################################################
        # IRGN - TV Reco ######################################################
        #######################################################################
        if "TV" in args.reg or args.reg == 'all':
            result_tv = []
            opt.model = model
            ###################################################################
            # IRGN Params #####################################################
            ###################################################################
            opt.irgn_par = utils.read_config(args.config, "3D_TV")
            opt.execute(TV=1, imagespace=args.imagespace)
            result_tv.append(opt.result)
            plt.close('all')
            res_tv = opt.gn_res
            res_tv = np.array(res_tv)/(opt.irgn_par["lambd"]*NSlice)
        del opt
###############################################################################
# New .hdf5 save files ########################################################
###############################################################################
    outdir = time.strftime("%Y-%m-%d  %H-%M-%S_MRI_"+args.reg+"_"+args.type+"_"
                           + fname[:-3])
    if not os.path.exists('./output'):
        os.makedirs('./output')
    if not os.path.exists('./output/' + outdir):
        os.makedirs("output/" + outdir)
    cwd = os.getcwd()
    os.chdir("output/" + outdir)
    f = h5py.File("output_" + fname, "w")
    f.create_dataset("images_ifft_", images.shape, dtype=DTYPE, data=images)
    if "TGV" in args.reg or args.reg == 'all':
        for i in range(len(result_tgv)):
            f.create_dataset("tgv_full_result_"+str(i), result_tgv[i].shape,
                             dtype=DTYPE, data=result_tgv[i])
            f.attrs['res_tgv'+str(i)] = res_tgv
            for j in range(len(model.uk_scale)):
                model.uk_scale[j] = 1
            image_final = model.execute_forward(result_tgv[i][-1, ...])
            f.create_dataset("sim_images_"+str(i), image_final.shape,
                             dtype=DTYPE, data=image_final)
    if "TV" in args.reg or args.reg == 'all':
        for i in range(len(result_tv)):
            f.create_dataset("tv_full_result_"+str(i), result_tv[i].shape,
                             dtype=DTYPE, data=result_tv[i])
            f.attrs['res_tv'] = res_tv

    if "imagespace" in args.reg or args.reg == 'all':
        for i in range(len(result_tgv)):
            f.create_dataset("imagespace_full_result_"+str(i),
                             result_tgv[i].shape,
                             dtype=DTYPE, data=result_tgv[i])
            f.attrs['imagespace_tgv'] = res_tgv
        f.flush()
    f.attrs['data_norm'] = dscale
    f.close()
    os.chdir(cwd)


def main(recon_type='3D', reg_type='TGV', slices=1, trafo=1, streamed=0,
         par_slices=1, data='', model='VFA', config='default', imagespace=0,
         OCL_GPU=1, devices=0, dz=1):
    """
    Start a 3D model based reconstruction. Data can also be selected at
    start up.

    Parameters
    ----------
    recon_type
        3D (2D currently not supported but 3D works on one slice also)
    reg_type
        TGV or TV, defaults to TGV
    slices
        The number of slices to reconsturct. Slices are picked symmetrically
        from the volume center. Pass -1 to select all slices available.
        Defaults to 1
    trafo
      Choos between Radial (1) or Cartesian (0) FFT
    streamed
      Toggle between streaming slices to the GPU (1) or computing
      everything with a single memory transfer (0). Defaults to 0
    par_slices
      Number of slices per streamed package. Volume devided by GPU's and
      par_slices must be an even number! Defaults to 1
    data
      The path to the .h5 file containing the data to reconstruct.
      If left empty, a GUI will open and asks for data file selection. This
      is also the default behaviour.
    model
      The name of the model which should be used to fit the data. Defaults to
      'VFA'. A path to your own model file can be passed. See the Model Class
      for further information on how to setup your own model.
    config
      The path to the confi gfile used for the IRGN reconstruction. If
      not specified the default config file will be used. If no default
      config file is present in the current working directory one will be
      generated.
    imagespace
      Select between fitting in imagespace (1) or in k-space (0). Defaults to 0
    OCL_GPU
      Select between GPU (1) or CPU (0) OpenCL devices. Defaults to GPU
      CAVE: CPU FFT not working.
    devices
      The device ID of device(s) to use for streaming/reconstruction
    dz
      Ratio of physical Z to X/Y dimension. X/Y is assumed to be isotropic.
    """
    parser = argparse.ArgumentParser(description="T1 quantification from VFA "
                                                 "data. By default runs 3D "
                                                 "regularization for TGV and "
                                                 "TV.")
    parser.add_argument(
      '--recon_type', default=recon_type, dest='type',
      help='Choose reconstruction type (currently only 3D)')
    parser.add_argument(
      '--reg_type', default=reg_type, dest='reg',
      help="Choose regularization type (default: TGV) "
           "options are: TGV, TV, all")
    parser.add_argument(
      '--slices', default=slices, dest='slices', type=int,
      help="Number of reconstructed slices (default=40). "
           "Symmetrical around the center slice.")
    parser.add_argument(
      '--trafo', default=trafo, dest='trafo', type=int,
      help='Choos between radial (1, default) and Cartesian (0) sampling. ')
    parser.add_argument(
      '--streamed', default=streamed, dest='streamed', type=int,
      help='Enable streaming of large data arrays (e.g. >10 slices).')
    parser.add_argument(
      '--par_slices', default=par_slices, dest='par_slices', type=int,
      help='number of slices per package. Volume devided by GPU\'s and'
           ' par_slices must be an even number!')
    parser.add_argument(
      '--data', default=data, dest='file',
      help="Full path to input data. "
           "If not provided, a file dialog will open.")
    parser.add_argument(
      '--model', default=model, dest='sig_model',
      help="Name of the signal model to use. Defaults to VFA. "
           "Please put your signal model file in the Model subfolder.")
    parser.add_argument(
      '--config', default=config, dest='config',
      help="Name of config file to use (assumed to be in the same folder). "
           "If not specified, use default parameters.")
    parser.add_argument(
      '--imagespace', default=imagespace, dest='imagespace', type=int,
      help="Select if Reco is performed on images (1) or on kspace (0) data. "
           "Defaults to 0")
    parser.add_argument(
      '--OCL_GPU', default=OCL_GPU, dest='use_GPU', type=int,
      help="Select if CPU or GPU should be used as OpenCL platform. "
           "Defaults to GPU (1). CAVE: CPU FFT not working")
    parser.add_argument(
      '--devices', default=devices, dest='devices', type=int,
      help="Device ID of device(s) to use for streaming. "
           "-1 selects all available devices", nargs='*')
    parser.add_argument(
      '--dz', default=dz, dest='dz', type=float,
      help="Ratio of physical Z to X/Y dimension. "
           "X/Y is assumed to be isotropic. Defaults to 1")
    args = parser.parse_args()
    start_recon(args)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="T1 quantification from VFA "
                                                 "data. By default runs 3D "
                                                 "regularization for TGV and "
                                                 "TV.")
    parser.add_argument(
      '--recon_type', default='3D', dest='type',
      help='Choose reconstruction type (currently only 3D)')
    parser.add_argument(
      '--reg_type', default='TGV', dest='reg',
      help="Choose regularization type (default: TGV) "
           "options are: TGV, TV, all")
    parser.add_argument(
      '--slices', default=1, dest='slices', type=int,
      help="Number of reconstructed slices (default=40). "
           "Symmetrical around the center slice.")
    parser.add_argument(
      '--trafo', default=1, dest='trafo', type=int,
      help='Choos between radial (1, default) and Cartesian (0) sampling. ')
    parser.add_argument(
      '--streamed', default=0, dest='streamed', type=int,
      help='Enable streaming of large data arrays (e.g. >10 slices).')
    parser.add_argument(
      '--par_slices', default=1, dest='par_slices', type=int,
      help='number of slices per package. Volume devided by GPU\'s and'
           ' par_slices must be an even number!')
    parser.add_argument(
      '--data', default='', dest='file',
      help="Full path to input data. "
           "If not provided, a file dialog will open.")
    parser.add_argument(
      '--model', default='VFA', dest='sig_model',
      help='Name of the signal model to use. Defaults to VFA. \
 Please put your signal model file in the Model subfolder.')
    parser.add_argument(
      '--config', default='default', dest='config',
      help='Name of config file to use (assumed to be in the same folder). \
 If not specified, use default parameters.')
    parser.add_argument(
      '--imagespace', default=0, dest='imagespace', type=int,
      help="Select if Reco is performed on images (1) or on kspace (0) data. "
           "Defaults to 0")
    parser.add_argument(
      '--imagespace', default=0, dest='imagespace', type=int,
      help='Select if Reco is performed on images (1) or on kspace (0) data. \
 Defaults to 0')
    parser.add_argument(
      '--devices', default=0, dest='devices', type=int,
      help="Device ID of device(s) to use for streaming. "
           "-1 selects all available devices", nargs='*')
    parser.add_argument(
      '--dz', default=1, dest='dz', type=float,
      help="Ratio of physical Z to X/Y dimension. "
           "X/Y is assumed to be isotropic. Defaults to  1")
    args = parser.parse_args()
    start_recon(args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2019 Oliver Maier

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np

import pyopencl as cl
import pyopencl.array as clarray
from gpyfft.fft import FFT
from pkg_resources import resource_filename
from pyqmri._helper_fun._calckbkernel import calckbkernel


class Program(object):
    def __init__(self, ctx, code):
        self._cl_prg = cl.Program(ctx, code)
        self._cl_prg.build("-cl-mad-enable -cl-fast-relaxed-math")
        self._cl_kernels = self._cl_prg.all_kernels()
        for kernel in self._cl_kernels:
            self.__dict__[kernel.function_name] = kernel


class PyOpenCLNUFFT:
    def __init__(
            self,
            ctx,
            queue,
            par,
            kwidth=5,
            overgridfactor=2,
            fft_dim=(
                1,
                2),
            klength=200,
            DTYPE=np.complex64,
            DTYPE_real=np.float32,
            radial=True):
        self.radial = radial
        self.DTYPE = DTYPE
        self.DTYPE_real = DTYPE_real
        self.fft_shape = (
            par["NScan"] *
            par["NC"] *
            par["NSlice"],
            par["N"],
            par["N"])
        self.traj = par["traj"]
        self.dcf = par["dcf"]
        self.ctx = ctx
        self.queue = queue
        if self.radial:
            (self.kerneltable, self.kerneltable_FT, self.u) = calckbkernel(
                kwidth, overgridfactor, par["N"], klength)
            self.kernelpoints = self.kerneltable.size
            self.overgridfactor = overgridfactor
            self.fft_scale = DTYPE_real(
                np.sqrt(np.prod(self.fft_shape[fft_dim[0]:])))
            self.deapo = 1 / self.kerneltable_FT.astype(DTYPE_real)
            self.kwidth = kwidth / 2
            self.cl_kerneltable = cl.Buffer(
                self.ctx,
                cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                hostbuf=self.kerneltable.astype(DTYPE_real).data)
            self.deapo_cl = cl.Buffer(
                self.ctx,
                cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                hostbuf=self.deapo.data)
            self.dcf = clarray.to_device(self.queue, self.dcf)
            self.traj = clarray.to_device(self.queue, self.traj)
            self.tmp_fft_array = (
                clarray.empty(
                    self.queue,
                    (self.fft_shape),
                    dtype=DTYPE))
            self.check = np.ones(par["N"], dtype=DTYPE_real)
            self.check[1::2] = -1
            self.check = clarray.to_device(self.queue, self.check)
            self.par_fft = int(self.fft_shape[0] / par["NScan"])
            self.fft = FFT(ctx, queue, self.tmp_fft_array[
                0:int(self.fft_shape[0] / par["NScan"]), ...],
                           out_array=self.tmp_fft_array[
                               0:int(self.fft_shape[0] / par["NScan"]), ...],
                           axes=fft_dim)
            self.gridsize = par["N"]
            self.fwd_NUFFT = self.NUFFT
            self.adj_NUFFT = self.NUFFTH
        else:
            self.fwd_NUFFT = self.FFT
            self.adj_NUFFT = self.FFTH
            self.fft_scale = DTYPE_real(
                np.sqrt(np.prod(self.fft_shape[fft_dim[0]:])))
            self.tmp_fft_array = (
                clarray.empty(
                    self.queue,
                    self.fft_shape,
                    dtype=DTYPE))
            self.fft_shape = self.fft_shape
            self.par_fft = int(self.fft_shape[0] / par["NScan"])
            self.mask = clarray.to_device(self.queue, par["mask"])
            self.fft = FFT(ctx, queue, self.tmp_fft_array[
                0:int(self.fft_shape[0] / par["NScan"]), ...],
                           out_array=self.tmp_fft_array[
                               0:int(self.fft_shape[0] / par["NScan"]), ...],
                           axes=fft_dim)
        print('Using single precission')
        self.prg = Program(
            self.ctx,
            open(
              resource_filename(
                'pyqmri',
                'kernels/OpenCL_gridding_single.c')).read())

    def __del__(self):
        if self.radial:
            del self.traj
            del self.dcf
            del self.tmp_fft_array
            del self.cl_kerneltable
            del self.deapo_cl
            del self.check
            del self.queue
            del self.ctx
        else:
            del self.tmp_fft_array
            del self.queue
            del self.ctx

    def NUFFTH(self, sg, s, wait_for=[]):
        # Zero tmp arrays
        self.tmp_fft_array.add_event(
            self.prg.zero_tmp(
                self.queue,
                (self.tmp_fft_array.size,
                 ),
                None,
                self.tmp_fft_array.data,
                wait_for=s.events +
                sg.events +
                self.tmp_fft_array.events +
                wait_for))
        # Grid k-space
        self.tmp_fft_array.add_event(
            self.prg.grid_lut(
                self.queue,
                (s.shape[0], s.shape[1] * s.shape[2],
                 s.shape[-2] * self.gridsize),
                None,
                self.tmp_fft_array.data,
                s.data,
                self.traj.data,
                np.int32(self.gridsize),
                self.DTYPE_real(self.kwidth / self.gridsize),
                self.dcf.data,
                self.cl_kerneltable,
                np.int32(self.kernelpoints),
                wait_for=(wait_for + sg.events +
                          s.events + self.tmp_fft_array.events)))
        # FFT
        self.tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self.tmp_fft_array.data,
                self.check.data))
        for j in range(s.shape[0]):
            self.tmp_fft_array.add_event(
                self.fft.enqueue_arrays(
                    data=self.tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    result=self.tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    forward=False)[0])
        self.tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self.tmp_fft_array.data,
                self.check.data))
        return self.prg.deapo_adj(
                  self.queue,
                  (sg.shape[0] * sg.shape[1] *
                   sg.shape[2], sg.shape[3], sg.shape[4]),
                  None,
                  sg.data,
                  self.tmp_fft_array.data,
                  self.deapo_cl,
                  np.int32(self.tmp_fft_array.shape[-1]),
                  self.DTYPE_real(self.fft_scale),
                  self.DTYPE_real(self.overgridfactor),
                  wait_for=(wait_for + sg.events + s.events +
                            self.tmp_fft_array.events))

    def NUFFT(self, s, sg, wait_for=[]):
        # Zero tmp arrays
        self.tmp_fft_array.add_event(
            self.prg.zero_tmp(
                self.queue,
                (self.tmp_fft_array.size,
                 ),
                None,
                self.tmp_fft_array.data,
                wait_for=s.events +
                sg.events +
                self.tmp_fft_array.events +
                wait_for))
        # Deapodization and Scaling
        self.tmp_fft_array.add_event(
            self.prg.deapo_fwd(
                self.queue,
                (sg.shape[0] * sg.shape[1] * sg.shape[2],
                 sg.shape[3], sg.shape[4]),
                None,
                self.tmp_fft_array.data,
                sg.data,
                self.deapo_cl,
                np.int32(self.tmp_fft_array.shape[-1]),
                self.DTYPE_real(1 / self.fft_scale),
                self.DTYPE_real(self.overgridfactor),
                wait_for=wait_for + sg.events + self.tmp_fft_array.events))
        # FFT
        self.tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self.tmp_fft_array.data,
                self.check.data))
        for j in range(s.shape[0]):
            self.tmp_fft_array.add_event(
                self.fft.enqueue_arrays(
                    data=self.tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    result=self.tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    forward=True)[0])
        self.tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self.tmp_fft_array.data,
                self.check.data))
        # Resample on Spoke
        return self.prg.invgrid_lut(
            self.queue,
            (s.shape[0], s.shape[1] * s.shape[2], s.shape[-2] * self.gridsize),
            None,
            s.data,
            self.tmp_fft_array.data,
            self.traj.data,
            np.int32(self.gridsize),
            self.DTYPE_real(self.kwidth / self.gridsize),
            self.dcf.data,
            self.cl_kerneltable,
            np.int32(self.kernelpoints),
            wait_for=s.events + wait_for + self.tmp_fft_array.events)

    def FFTH(self, sg, s, wait_for=[]):

        s.add_event(
            self.prg.masking(
                self.queue,
                (s.size,
                 ),
                None,
                s.data,
                self.mask.data,
                wait_for=s.events))
        self.tmp_fft_array.add_event(
            self.prg.copy(
                self.queue,
                (s.size,
                 ),
                None,
                self.tmp_fft_array.data,
                (s).data,
                self.DTYPE_real(1),
                wait_for=s.events))

        for j in range(s.shape[0]):
            self.tmp_fft_array.add_event(
                self.fft.enqueue_arrays(
                    data=self.tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    result=self.tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    forward=False)[0])
        return (
            self.prg.copy(
                self.queue,
                (sg.size,
                 ),
                None,
                sg.data,
                self.tmp_fft_array.data,
                self.DTYPE_real(
                    self.fft_scale)))

    def FFT(self, s, sg, wait_for=[]):
        self.tmp_fft_array.add_event(
            self.prg.copy(
                self.queue,
                (sg.size,
                 ),
                None,
                self.tmp_fft_array.data,
                sg.data,
                self.DTYPE_real(
                    1 /
                    self.fft_scale)))

        for j in range(s.shape[0]):
            self.tmp_fft_array.add_event(
                self.fft.enqueue_arrays(
                    data=self.tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    result=self.tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    forward=True)[0])

        s.add_event(
            self.prg.copy(
                self.queue,
                (s.size,
                 ),
                None,
                s.data,
                self.tmp_fft_array.data,
                self.DTYPE_real(1),
                wait_for=self.tmp_fft_array.events))
        return (
            self.prg.masking(
                self.queue,
                (s.size,
                 ),
                None,
                s.data,
                self.mask.data,
                wait_for=s.events))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2019 Oliver Maier

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division

import numpy as np
import time
import sys
from pkg_resources import resource_filename
import pyopencl as cl
import pyopencl.array as clarray

import pyqmri._transforms._pyopencl_nufft as NUFFT

DTYPE = np.complex64
DTYPE_real = np.float32


class Program(object):
    def __init__(self, ctx, code):
        self._cl_prg = cl.Program(ctx, code)
        self._cl_prg.build("-cl-mad-enable -cl-fast-relaxed-math")
        self._cl_kernels = self._cl_prg.all_kernels()
        for kernel in self._cl_kernels:
            self.__dict__[kernel.function_name] = kernel


class ModelReco:
    def __init__(self, par, trafo=1, imagespace=False):
        self.C = par["C"]
        self.traj = par["traj"]
        self.unknowns_TGV = par["unknowns_TGV"]
        self.unknowns_H1 = par["unknowns_H1"]
        self.unknowns = par["unknowns"]
        self.NSlice = par["NSlice"]
        self.NScan = par["NScan"]
        self.dimX = par["dimX"]
        self.dimY = par["dimY"]
        self.NC = par["NC"]
        self.fval_old = 0
        self.fval = 0
        self.fval_init = 0
        self.SNR_est = par["SNR_est"]
        self.ctx = par["ctx"][0]
        self.queue = par["queue"][0]
        self.ratio = clarray.to_device(
            self.queue,
            (1 /
             self.unknowns *
             np.ones(
                 self.unknowns)).astype(
                dtype=DTYPE_real))
        self.gn_res = []
        self.N = par["N"]
        self.Nproj = par["Nproj"]
        self.dz = 1
        self.tmp_result = clarray.empty(
            self.queue, (self.NScan, self.NSlice, self.dimY, self.dimX),
            DTYPE, "C")
        if imagespace:
            self.operator_forward = self.operator_forward_imagespace
            self.operator_adjoint = self.operator_adjoint_imagespace
        else:
            self.coil_buf = cl.Buffer(self.ctx,
                                      cl.mem_flags.READ_ONLY |
                                      cl.mem_flags.COPY_HOST_PTR,
                                      hostbuf=self.C.data)
            self.operator_forward = self.operator_forward_kspace
            self.operator_adjoint = self.operator_adjoint_kspace
            self.tmp_sino = clarray.empty(
                self.queue,
                (self.NScan, self.NC, self.NSlice, self.Nproj, self.N),
                DTYPE, "C")
            self.tmp_result = clarray.empty(
                self.queue,
                (self.NScan, self.NC, self.NSlice, self.dimY, self.dimX),
                DTYPE, "C")
            self.NUFFT = NUFFT.PyOpenCLNUFFT(self.ctx, self.queue,
                                             par, radial=trafo)

        self.prg = Program(
            self.ctx,
            open(resource_filename('pyqmri', 'kernels/OpenCL_Kernels.c')).read())

    def operator_forward_kspace(self, x, out=None, wait_for=[]):
        self.tmp_result.add_event(
            self.prg.operator_fwd(
                self.queue,
                (self.NSlice, self.dimY, self.dimX),
                None,
                self.tmp_result.data, x.data,
                self.coil_buf,
                self.grad_buf, np.int32(self.NC),
                np.int32(self.NScan),
                np.int32(self.unknowns),
                wait_for=self.tmp_result.events + x.events + wait_for))
        if out is not None:
            return self.NUFFT.fwd_NUFFT(
                out,
                self.tmp_result,
                wait_for=wait_for +
                self.tmp_result.events)
        else:
            self.tmp_sino.add_event(
                self.NUFFT.fwd_NUFFT(self.tmp_sino, self.tmp_result))
            return self.tmp_sino

    def operator_adjoint_kspace(self, x, out=None, z=None, wait_for=[]):
        self.tmp_result.add_event(
            self.NUFFT.adj_NUFFT(
                self.tmp_result, x, wait_for=wait_for + x.events))
        if out is not None and z is not None:
            return self.prg.update_Kyk1(
                self.queue, (self.NSlice, self.dimY, self.dimX), None,
                out.data, self.tmp_result.data, self.coil_buf,
                self.grad_buf, z.data, np.int32(self.NC),
                np.int32(self.NScan),
                self.ratio.data,
                np.int32(self.unknowns), np.float32(self.dz),
                wait_for=(self.tmp_result.events +
                          out.events + z.events + wait_for))
        elif out is not None and z is None:
            return self.prg.operator_ad(
                out.queue, (self.NSlice, self.dimY, self.dimX), None,
                out.data, self.tmp_result.data, self.coil_buf,
                self.grad_buf, np.int32(self.NC),
                np.int32(self.NScan),
                np.int32(self.unknowns),
                wait_for=wait_for + self.tmp_result.events + out.events)
        else:
            out = clarray.empty(
                self.queue, (self.unkowns, self.NSlice, self.dimY, self.dimX,),
                dtype=DTYPE)
            self.prg.operator_ad(
                out.queue, (self.NSlice, self.dimY, self.dimX), None,
                out.data, self.tmp_result.data, self.coil_buf,
                self.grad_buf, np.int32(self.NC),
                np.int32(self.NScan),
                np.int32(self.unknowns),
                wait_for=wait_for + self.tmp_result.events + out.events).wait()
            return out

    def operator_forward_imagespace(self, x, out=None, wait_for=[]):
        if out is not None:
            return self.prg.operator_fwd_imagespace(
                self.queue, (self.NSlice, self.dimY, self.dimX), None,
                out.data, x.data, self.grad_buf,
                np.int32(self.NScan),
                np.int32(self.unknowns),
                wait_for=x.events + out.events + wait_for)
        else:
            self.tmp_result.add_event(self.prg.operator_fwd_imagespace(
                self.queue, (self.NSlice, self.dimY, self.dimX), None,
                self.tmp_result.data, x.data, self.grad_buf,
                np.int32(self.NScan),
                np.int32(self.unknowns),
                wait_for=x.events + wait_for))
            return self.tmp_result

    def operator_adjoint_imagespace(self, x, out=None, z=None, wait_for=[]):
        if out is not None and z is not None:
            return self.prg.update_Kyk1_imagespace(
                self.queue, (self.NSlice, self.dimY, self.dimX), None,
                out.data, x.data, self.grad_buf, z.data,
                np.int32(self.NScan),
                self.ratio.data,
                np.int32(self.unknowns),
                np.float32(self.dz),
                wait_for=x.events + out.events + z.events + wait_for)
        elif out is not None and z is None:
            return self.prg.operator_ad_imagespace(
                out.queue, (self.NSlice, self.dimY, self.dimX), None,
                out.data, x.data, self.grad_buf,
                np.int32(self.NScan),
                np.int32(self.unknowns),
                wait_for=wait_for + x.events + out.events)
        else:
            out = clarray.empty(
                self.queue, (self.unkowns, self.NSlice, self.dimY, self.dimX,),
                dtype=DTYPE)
            self.prg.operator_ad_imagespace(
                out.queue, (self.NSlice, self.dimY, self.dimX), None,
                out.data, x.data, self.grad_buf,
                np.int32(self.NScan),
                np.int32(self.unknowns),
                wait_for=wait_for + x.events + out.events).wait()
            return out

    def f_grad(self, grad, u, wait_for=[]):
        return self.prg.gradient(
            self.queue, u.shape[1:], None, grad.data, u.data,
            np.int32(self.unknowns),
            self.ratio.data, np.float32(self.dz),
            wait_for=grad.events + u.events + wait_for)

    def bdiv(self, div, u, wait_for=[]):
        return self.prg.divergence(
            div.queue, u.shape[1:-1], None, div.data, u.data,
            np.int32(self.unknowns), self.ratio.data,
            np.float32(self.dz), wait_for=div.events + u.events + wait_for)

    def sym_grad(self, sym, w, wait_for=[]):
        return self.prg.sym_grad(
            self.queue, w.shape[1:-1], None, sym.data, w.data,
            np.int32(self.unknowns_TGV), np.float32(self.dz),
            wait_for=sym.events + w.events + wait_for)

    def sym_bdiv(self, div, u, wait_for=[]):
        return self.prg.sym_divergence(
            self.queue, u.shape[1:-1], None, div.data, u.data,
            np.int32(self.unknowns_TGV), np.float32(self.dz),
            wait_for=div.events + u.events + wait_for)

    def update_Kyk2(self, div, u, z, wait_for=[]):
        return self.prg.update_Kyk2(
            self.queue, u.shape[1:-1], None, div.data, u.data, z.data,
            np.int32(self.unknowns_TGV), np.float32(self.dz),
            wait_for=div.events + u.events + z.events + wait_for)

    def update_primal(self, x_new, x, Kyk, xk, tau, delta, wait_for=[]):
        return self.prg.update_primal(
            self.queue, x.shape[1:], None, x_new.data, x.data, Kyk.data,
            xk.data, np.float32(tau),
            np.float32(tau / delta), np.float32(1 / (1 + tau / delta)),
            self.min_const.data, self.max_const.data,
            self.real_const.data, np.int32(self.unknowns),
            wait_for=(x_new.events + x.events +
                      Kyk.events + xk.events + wait_for))

    def update_z1(self, z_new, z, gx, gx_, vx, vx_,
                  sigma, theta, alpha, omega, wait_for=[]):
        return self.prg.update_z1(
            self.queue, z.shape[1:-1], None, z_new.data, z.data, gx.data,
            gx_.data, vx.data, vx_.data, np.float32(sigma), np.float32(theta),
            np.float32(1 / alpha), np.int32(self.unknowns_TGV),
            np.int32(self.unknowns_H1), np.float32(1 / (1 + sigma / omega)),
            wait_for=(
                z_new.events + z.events + gx.events +
                gx_.events + vx.events + vx_.events + wait_for))

    def update_z1_tv(self, z_new, z, gx, gx_,
                     sigma, theta, alpha, omega, wait_for=[]):
        return self.prg.update_z1_tv(
            self.queue, z.shape[1:-1], None, z_new.data, z.data, gx.data,
            gx_.data, np.float32(sigma), np.float32(theta),
            np.float32(1 / alpha), np.int32(self.unknowns_TGV),
            np.int32(self.unknowns_H1), np.float32(1 / (1 + sigma / omega)),
            wait_for=(
                z_new.events + z.events + gx.events + gx_.events + wait_for))

    def update_z2(self, z_new, z, gx, gx_, sigma, theta, beta, wait_for=[]):
        return self.prg.update_z2(
            self.queue, z.shape[1:-1], None, z_new.data, z.data,
            gx.data, gx_.data, np.float32(sigma), np.float32(theta),
            np.float32(1 / beta), np.int32(self.unknowns_TGV),
            wait_for=(
                z_new.events + z.events + gx.events + gx_.events + wait_for))

    def update_r(self, r_new, r, A, A_, res, sigma, theta, lambd, wait_for=[]):
        return self.prg.update_r(
            self.queue, (r.size,), None, r_new.data, r.data, A.data, A_.data,
            res.data, np.float32(sigma), np.float32(theta),
            np.float32(1 / (1 + sigma / lambd)),
            wait_for=r_new.events + r.events + A.events + A_.events + wait_for)

    def update_v(self, v_new, v, Kyk2, tau, wait_for=[]):
        return self.prg.update_v(
            self.queue, (v[..., 0].size,), None, v_new.data, v.data,
            Kyk2.data, np.float32(tau),
            wait_for=v_new.events + v.events + Kyk2.events + wait_for)

    def update_primal_explicit(self, x_new, x, Kyk, xk, ATd,
                               tau, delta, lambd, wait_for=[]):
        return self.prg.update_primal_explicit(
            self.queue, x.shape[1:], None, x_new.data, x.data, Kyk.data,
            xk.data, ATd.data, np.float32(tau),
            np.float32(1 / delta), np.float32(lambd), self.min_const.data,
            self.max_const.data,
            self.real_const.data, np.int32(self.unknowns),
            wait_for=(
                x_new.events + x.events + Kyk.events +
                xk.events + ATd.events + wait_for))

    def eval_const(self):
        num_const = (len(self.model.constraints))
        self.min_const = np.zeros((num_const), dtype=np.float32)
        self.max_const = np.zeros((num_const), dtype=np.float32)
        self.real_const = np.zeros((num_const), dtype=np.int32)
        for j in range(num_const):
            self.min_const[j] = np.float32(self.model.constraints[j].min)
            self.max_const[j] = np.float32(self.model.constraints[j].max)
            self.real_const[j] = np.int32(self.model.constraints[j].real)
        self.min_const = clarray.to_device(self.queue, self.min_const)
        self.max_const = clarray.to_device(self.queue, self.max_const)
        self.real_const = clarray.to_device(self.queue, self.real_const)

###############################################################################
# Scale before gradient #######################################################
###############################################################################
    def set_scale(self, x):
        x = clarray.to_device(self.queue, x)
        grad = clarray.to_device(self.queue, np.zeros_like(self.z1))
        grad.add_event(
            self.f_grad(
                grad,
                x,
                wait_for=grad.events +
                x.events))
        x = x.get()
        grad = grad.get()
        scale = np.reshape(
            x, (self.unknowns, self.NSlice * self.dimY * self.dimX))
        grad = np.reshape(
            grad, (self.unknowns, self.NSlice * self.dimY * self.dimX * 4))
        print("Diff between x: ", np.linalg.norm(scale, axis=-1))
        print("Diff between grad x: ", np.linalg.norm(grad, axis=-1))
        scale = np.linalg.norm(grad, axis=-1)
        scale = 1/scale
        scale[~np.isfinite(scale)] = 1
        sum_scale = np.linalg.norm(
            scale[:self.unknowns_TGV])/(1000/np.sqrt(self.NSlice))
        for j in range(x.shape[0])[:self.unknowns_TGV]:
            self.ratio[j] = scale[j] / sum_scale
        sum_scale = np.linalg.norm(
            scale[self.unknowns_TGV:])/(1000)
        for j in range(x.shape[0])[self.unknowns_TGV:]:
            self.ratio[j] = scale[j] / sum_scale
        print("Ratio: ", self.ratio)

###############################################################################
# Start a 3D Reconstruction, set TV to True to perform TV instead of TGV#######
# Precompute Model and Gradient values for xk #################################
# Call inner optimization #####################################################
# input: bool to switch between TV (1) and TGV (0) regularization #############
# output: optimal value of x ##################################################
###############################################################################
    def execute_3D(self, TV=0):
        self.FT = self.NUFFT.fwd_NUFFT
        self.FTH = self.NUFFT.adj_NUFFT

        iters = self.irgn_par["start_iters"]
        self.r = np.zeros_like(self.data, dtype=DTYPE)
        self.z1 = np.zeros(
            ([self.unknowns, self.NSlice, self.dimY, self.dimX, 4]),
            dtype=DTYPE)

        self.result = np.zeros(
            (self.irgn_par["max_gn_it"] + 1, self.unknowns, self.NSlice,
             self.dimY, self.dimX),
            dtype=DTYPE)
        self.result[0, ...] = np.copy(self.model.guess)

        result = np.copy(self.model.guess)

        if TV == 1:
            self.tau = np.float32(1 / np.sqrt(8))
            self.beta_line = 400
            self.theta_line = np.float32(1.0)
            pass
        elif TV == 0:
            L = np.float32(0.5 * (18.0 + np.sqrt(33)))
            self.tau = np.float32(1 / np.sqrt(L))
            self.beta_line = 400
            self.theta_line = np.float32(1.0)
            self.v = np.zeros(
                ([self.unknowns_TGV, self.NSlice, self.dimY, self.dimX, 4]),
                dtype=DTYPE)
            self.z2 = np.zeros(
                ([self.unknowns_TGV, self.NSlice, self.dimY, self.dimX, 8]),
                dtype=DTYPE)
        else:
            print("Not implemented")
            return

        for i in range(self.irgn_par["max_gn_it"]):
            start = time.time()
            self.grad_x = np.nan_to_num(self.model.execute_gradient(result))
            scale = np.reshape(
                self.grad_x,
                (self.unknowns,
                 self.NScan * self.NSlice * self.dimY * self.dimX))
            scale = np.linalg.norm(scale, axis=-1)
            print("Initial norm of the model Gradient: \n", scale)
            scale = 1e3 / np.sqrt(self.unknowns) / scale
            print("Scalefactor of the model Gradient: \n", scale)
            if not np.mod(i, 1):
                for uk in range(self.unknowns):
                    self.model.constraints[uk].update(scale[uk])
                    result[uk, ...] *= self.model.uk_scale[uk]
                    self.grad_x[uk] /= self.model.uk_scale[uk]
                    self.model.uk_scale[uk] *= scale[uk]
                    result[uk, ...] /= self.model.uk_scale[uk]
                    self.grad_x[uk] *= self.model.uk_scale[uk]
            scale = np.reshape(
                self.grad_x,
                (self.unknowns,
                 self.NScan * self.NSlice * self.dimY * self.dimX))
            scale = np.linalg.norm(scale, axis=-1)
            print("Scale of the model Gradient: \n", scale)

            self.set_scale(result)

            self.step_val = np.nan_to_num(self.model.execute_forward(result))
            self.grad_buf = cl.Buffer(self.queue.context,
                                      cl.mem_flags.READ_ONLY |
                                      cl.mem_flags.COPY_HOST_PTR,
                                      hostbuf=self.grad_x.data)

            self.irgn_par["delta_max"] = self.delta_max / \
                                         (1e3) * np.linalg.norm(result)
            self.irgn_par["delta"] = np.minimum(
                self.delta /
                (1e3)*np.linalg.norm(result)*self.irgn_par["delta_inc"]**i,
                self.irgn_par["delta_max"])

            result = self.irgn_solve_3D(result, iters, self.data, i, TV)
            self.result[i + 1, ...] = self.model.rescale(result)

            iters = np.fmin(iters * 2, self.irgn_par["max_iters"])
            self.irgn_par["gamma"] = np.maximum(
                self.irgn_par["gamma"] * self.irgn_par["gamma_dec"],
                self.irgn_par["gamma_min"])
            self.irgn_par["omega"] = np.maximum(
                self.irgn_par["omega"] * self.irgn_par["omega_dec"],
                self.irgn_par["omega_min"])

            end = time.time() - start
            self.gn_res.append(self.fval)
            print("-" * 75)
            print("GN-Iter: %d  Elapsed time: %f seconds" % (i, end))
            print("-" * 75)
            if np.abs(self.fval_old - self.fval) / self.fval_init < \
               self.irgn_par["tol"]:
                print("Terminated at GN-iteration %d because "
                      "the energy decrease was less than %.3e" %
                      (i,  np.abs(self.fval_old - self.fval) / self.fval_init))
                self.calc_residual_kspace(result, self.data, i+1, TV)
                break
            self.fval_old = self.fval
        self.calc_residual_kspace(result, self.data, i+1, TV)

###############################################################################
# Precompute constant terms of the GN linearization step ######################
# input: linearization point x ################################################
########## numeber of innner iterations iters #################################
########## Data ###############################################################
########## bool to switch between TV (1) and TGV (0) regularization ###########
# output: optimal value of x for the inner GN step ############################
###############################################################################
###############################################################################
    def irgn_solve_3D(self, x, iters, data, GN_it, TV=0):
        x = clarray.to_device(self.queue, np.require(x, requirements="C"))
        b = clarray.empty(self.queue, data.shape, dtype=DTYPE)
        self.FT(b, clarray.to_device(
            self.queue, self.step_val[:, None, ...] * self.C)).wait()
        res = data - b.get() + self.operator_forward(x).get()
        x = x.get()
        self.calc_residual_kspace(x, data, GN_it, TV)

        if TV == 1:
            x = self.tv_solve_3D(x, res, iters)
        elif TV == 0:
            x = self.tgv_solve_3D(x, res, iters)
        return x

    def calc_residual_kspace(self, x, data, GN_it, TV=0):
        x = clarray.to_device(self.queue, np.require(x, requirements="C"))
        b = clarray.empty(self.queue, data.shape, dtype=DTYPE)
        if TV == 1:
            grad = clarray.to_device(self.queue, np.zeros_like(self.z1))
            grad.add_event(
                self.f_grad(
                    grad,
                    x,
                    wait_for=grad.events +
                    x.events))
            x = x.get()
            self.FT(b, clarray.to_device(
                self.queue,
                (self.step_val[:, None, ...] *
                 self.C))).wait()
            grad = grad.get()
            self.fval = (self.irgn_par["lambd"] / 2 *
                         np.linalg.norm(data - b.get())**2 +
                         self.irgn_par["gamma"] *
                         np.sum(np.abs(grad[:self.unknowns_TGV])) +
                         self.irgn_par["omega"] / 2 *
                         np.linalg.norm(grad[self.unknowns_TGV:])**2)
            del grad, b
        elif TV == 0:
            v = clarray.to_device(self.queue, self.v)
            grad = clarray.to_device(self.queue, np.zeros_like(self.z1))
            sym_grad = clarray.to_device(self.queue, np.zeros_like(self.z2))
            grad.add_event(
                self.f_grad(
                    grad,
                    x,
                    wait_for=grad.events +
                    x.events))
            sym_grad.add_event(
                self.sym_grad(
                    sym_grad,
                    v,
                    wait_for=sym_grad.events +
                    v.events))
            x = x.get()
            grad = grad.get()
            self.FT(
                b,
                clarray.to_device(
                    self.queue,
                    self.step_val[:, None, ...] *
                    self.C)).wait()
            self.fval = (self.irgn_par["lambd"] / 2 *
                         np.linalg.norm(data - b.get())**2 +
                         self.irgn_par["gamma"] *
                         np.sum(np.abs(grad[:self.unknowns_TGV] - self.v)) +
                         self.irgn_par["gamma"] * (2) *
                         np.sum(np.abs(sym_grad.get())) +
                         self.irgn_par["omega"] / 2 *
                         np.linalg.norm(grad[self.unknowns_TGV:])**2)
            del grad, sym_grad, v, b
        if GN_it == 0:
            self.fval_init = self.fval
        print("-" * 75)
        print("Function value at GN-Step %i: %f" %
              (GN_it, 1e3*self.fval / self.fval_init))
        print("-" * 75)

###############################################################################
# Start a 3D Reconstruction, set TV to True to perform TV instead of TGV#######
# Precompute Model and Gradient values for xk #################################
# Call inner optimization #####################################################
# input: bool to switch between TV (1) and TGV (0) regularization #############
# output: optimal value of x ##################################################
###############################################################################
    def execute_3D_imagespace(self, TV=0):
        iters = self.irgn_par["start_iters"]
        self.NC = 1
        self.N = self.dimX
        self.Nproj = self.dimY

        self.r = np.zeros_like(self.data, dtype=DTYPE)
        self.z1 = np.zeros(
            ([self.unknowns, self.NSlice, self.dimY, self.dimX, 4]),
            dtype=DTYPE)

        self.result = np.zeros(
            (self.irgn_par["max_gn_it"] + 1, self.unknowns,
             self.NSlice, self.dimY, self.dimX),
            dtype=DTYPE)
        self.result[0, ...] = self.model.rescale(self.model.guess)

        result = np.copy(self.model.guess)

        if TV == 1:
            self.tau = np.float32(1 / np.sqrt(8))
            self.beta_line = 1
            self.theta_line = np.float32(1.0)
            pass
        elif TV == 0:
            L = np.float32(0.5 * (18.0 + np.sqrt(33)))
            self.tau = np.float32(1 / np.sqrt(L))
            self.beta_line = 1
            self.theta_line = np.float32(1.0)
            self.v = np.zeros(
                ([self.unknowns_TGV, self.NSlice, self.dimY, self.dimX, 4]),
                dtype=DTYPE)
            self.z2 = np.zeros(
                ([self.unknowns_TGV, self.NSlice, self.dimY, self.dimX, 8]),
                dtype=DTYPE)
        else:
            print("Not implemented")
            return

        for i in range(self.irgn_par["max_gn_it"]):
            start = time.time()
            self.grad_x = np.nan_to_num(self.model.execute_gradient(result))
            scale = np.reshape(
                self.grad_x,
                (self.unknowns,
                 self.NScan * self.NSlice * self.dimY * self.dimX))
            scale = np.linalg.norm(scale, axis=-1)
            print("Initial norm of the model Gradient: \n", scale)
            scale = 1e3 / np.sqrt(self.unknowns) / scale
            print("Scalefactor of the model Gradient: \n", scale)
            if not np.mod(i, 1):
                for uk in range(self.unknowns):
                    self.model.constraints[uk].update(scale[uk])
                    result[uk, ...] *= self.model.uk_scale[uk]
                    self.grad_x[uk] /= self.model.uk_scale[uk]
                    self.model.uk_scale[uk] *= scale[uk]
                    result[uk, ...] /= self.model.uk_scale[uk]
                    self.grad_x[uk] *= self.model.uk_scale[uk]
            scale = np.reshape(
                self.grad_x,
                (self.unknowns,
                 self.NScan * self.NSlice * self.dimY * self.dimX))
            scale = np.linalg.norm(scale, axis=-1)
            print("Scale of the model Gradient: \n", scale)

            self.set_scale(result)

            self.step_val = np.nan_to_num(self.model.execute_forward(result))
            self.grad_buf = cl.Buffer(self.queue.context,
                                      cl.mem_flags.READ_ONLY |
                                      cl.mem_flags.COPY_HOST_PTR,
                                      hostbuf=self.grad_x.data)

            self.irgn_par["delta_max"] = self.delta_max / \
                                         (1e3) * np.linalg.norm(result)
            self.irgn_par["delta"] = np.minimum(
                self.delta /
                (1e3)*np.linalg.norm(result)*self.irgn_par["delta_inc"]**i,
                self.irgn_par["delta_max"])

            result = self.irgn_solve_3D_imagespace(result, iters,
                                                   self.data, i, TV)
            self.result[i + 1, ...] = self.model.rescale(result)

            iters = np.fmin(iters * 2, self.irgn_par["max_iters"])
            self.irgn_par["gamma"] = np.maximum(
                self.irgn_par["gamma"] * self.irgn_par["gamma_dec"],
                self.irgn_par["gamma_min"])
            self.irgn_par["omega"] = np.maximum(
                self.irgn_par["omega"] * self.irgn_par["omega_dec"],
                self.irgn_par["omega_min"])

            end = time.time() - start
            self.gn_res.append(self.fval)
            print("-" * 75)
            print("GN-Iter: %d  Elapsed time: %f seconds" % (i, end))
            print("-" * 75)
            if np.abs(self.fval_old - self.fval) / self.fval_init < \
               self.irgn_par["tol"]:
                print("Terminated at GN-iteration %d because the "
                      "relative energy decrease was less than %.3e" %
                      (i, np.abs(self.fval_old - self.fval) / self.fval_init))
                self.calc_residual_imagespace(result, self.data, i+1, TV)
                break
            self.fval_old = self.fval
        self.calc_residual_imagespace(result, self.data, i+1, TV)

###############################################################################
# Precompute constant terms of the GN linearization step ######################
# input: linearization point x ################################################
########## numeber of innner iterations iters #################################
########## Data ###############################################################
########## bool to switch between TV (1) and TGV (0) regularization ###########
# output: optimal value of x for the inner GN step ############################
###############################################################################
###############################################################################
    def irgn_solve_3D_imagespace(self, x, iters, data, GN_it, TV=0):
        x = clarray.to_device(self.queue, np.require(x, requirements="C"))
        res = data - self.step_val + self.operator_forward(x).get()
        x = x.get()

        self.calc_residual_imagespace(x, data, GN_it, TV)

        if TV == 1:
            x = self.tv_solve_3D(x, res, iters)
        elif TV == 0:
            x = self.tgv_solve_3D(x, res, iters)
        return x

    def calc_residual_imagespace(self, x, data, GN_it, TV=0):
        x = clarray.to_device(self.queue, np.require(x, requirements="C"))
        if TV == 1:
            grad = clarray.to_device(self.queue, np.zeros_like(self.z1))
            grad.add_event(
                self.f_grad(
                    grad,
                    x,
                    wait_for=grad.events +
                    x.events))
            x = x.get()
            grad = grad.get()
            self.fval = (
                self.irgn_par["lambd"] / 2 *
                np.linalg.norm(data - self.step_val)**2 +
                self.irgn_par["gamma"] *
                np.sum(np.abs(grad[:self.unknowns_TGV])) +
                self.irgn_par["omega"] / 2 *
                np.linalg.norm(grad[self.unknowns_TGV:])**2)
            del grad
        elif TV == 0:
            v = clarray.to_device(self.queue, self.v)
            grad = clarray.to_device(self.queue, np.zeros_like(self.z1))
            sym_grad = clarray.to_device(self.queue, np.zeros_like(self.z2))

            grad.add_event(self.f_grad(
                grad, x, wait_for=grad.events + x.events))
            sym_grad.add_event(self.sym_grad(
                sym_grad, v, wait_for=sym_grad.events + v.events))
            x = x.get()
            grad = grad.get()
            self.fval = (
                self.irgn_par["lambd"] / 2 *
                np.linalg.norm(data - self.step_val)**2 +
                self.irgn_par["gamma"] *
                np.sum(np.abs(grad[:self.unknowns_TGV] - self.v)) +
                self.irgn_par["gamma"] * (2) * np.sum(np.abs(sym_grad.get())) +
                self.irgn_par["omega"] / 2 *
                np.linalg.norm(grad[self.unknowns_TGV:])**2)
            del grad, sym_grad, v
        if GN_it == 0:
            self.fval_init = self.fval
        print("-" * 75)
        print("Function value at GN-Step %i: %f" %
              (GN_it, 1e3*self.fval / self.fval_init))
        print("-" * 75)

    def tgv_solve_3D(self, x, res, iters):
        alpha = self.irgn_par["gamma"]
        beta = self.irgn_par["gamma"] * 2

        tau = self.tau
        tau_new = np.float32(0)

        x = clarray.to_device(self.queue, x)
        xk = x.copy()
        x_new = clarray.empty_like(x)

        r = clarray.to_device(self.queue, np.zeros_like(self.r))
        r_new = clarray.empty_like(r)
        z1 = clarray.to_device(self.queue, np.zeros_like(self.z1))
        z1_new = clarray.empty_like(z1)
        z2 = clarray.to_device(self.queue, np.zeros_like(self.z2))
        z2_new = clarray.empty_like(z2)
        v = clarray.to_device(self.queue, np.zeros_like(self.v))
        v_new = clarray.empty_like(v)
        res = clarray.to_device(self.queue, res.astype(DTYPE))

        delta = self.irgn_par["delta"]
        omega = self.irgn_par["omega"]
        mu = 1 / delta

        theta_line = self.theta_line
        beta_line = self.beta_line
        beta_new = np.float32(0)
        mu_line = np.float32(0.5)
        delta_line = np.float32(1)
        ynorm = np.float32(0.0)
        lhs = np.float32(0.0)
        primal = np.float32(0.0)
        primal_new = np.float32(0)
        dual = np.float32(0.0)
        gap_init = np.float32(0.0)
        gap_old = np.float32(0.0)
        gap = np.float32(0.0)
        self.eval_const()

        Kyk1 = clarray.empty_like(x)
        Kyk1_new = clarray.empty_like(x)
        Kyk2 = clarray.empty_like(z1)
        Kyk2_new = clarray.empty_like(z1)
        gradx = clarray.empty_like(z1)
        gradx_xold = clarray.empty_like(z1)
        symgrad_v = clarray.empty_like(z2)
        symgrad_v_vold = clarray.empty_like(z2)
        Axold = clarray.empty_like(res)
        Ax = clarray.empty_like(res)

        Axold.add_event(self.operator_forward(x, Axold))
        Kyk1.add_event(self.operator_adjoint(r, Kyk1, z1))
        Kyk2.add_event(self.update_Kyk2(Kyk2, z2, z1))
        gradx_xold.add_event(self.f_grad(gradx_xold, x))
        symgrad_v_vold.add_event(self.sym_grad(symgrad_v_vold, v))

        for i in range(iters):
            x_new.add_event(self.update_primal(x_new, x, Kyk1, xk, tau, delta))
            v_new.add_event(self.update_v(v_new, v, Kyk2, tau))

            beta_new = beta_line * (1 + mu * tau)
            tau_new = tau * np.sqrt(beta_line / beta_new * (1 + theta_line))
            beta_line = beta_new

            gradx.add_event(self.f_grad(gradx, x_new))
            symgrad_v.add_event(self.sym_grad(symgrad_v, v_new))
            Ax.add_event(self.operator_forward(x_new, Ax))

            while True:
                theta_line = tau_new / tau
                z1_new.add_event(self.update_z1(
                    z1_new, z1, gradx, gradx_xold, v_new, v,
                    beta_line * tau_new, theta_line, alpha,
                    omega))
                z2_new.add_event(self.update_z2(
                    z2_new, z2, symgrad_v, symgrad_v_vold,
                    beta_line * tau_new, theta_line, beta))
                r_new.add_event(self.update_r(
                    r_new, r, Ax, Axold, res,
                    beta_line * tau_new, theta_line, self.irgn_par["lambd"]))
                Kyk1_new.add_event(self.operator_adjoint(
                    r_new, Kyk1_new, z1_new))
                Kyk2_new.add_event(self.update_Kyk2(Kyk2_new, z2_new, z1_new))

                ynorm = (
                    (clarray.vdot(r_new - r, r_new - r) +
                     clarray.vdot(z1_new - z1, z1_new - z1) +
                     clarray.vdot(z2_new - z2, z2_new - z2))**(1 / 2)).real
                lhs = np.sqrt(beta_line) * tau_new * (
                    (clarray.vdot(Kyk1_new - Kyk1, Kyk1_new - Kyk1) +
                     clarray.vdot(Kyk2_new - Kyk2, Kyk2_new - Kyk2))**(1 / 2)
                    ).real

                if lhs <= ynorm * delta_line:
                    break
                else:
                    tau_new = tau_new * mu_line

            (Kyk1, Kyk1_new, Kyk2, Kyk2_new, Axold, Ax, z1, z1_new,
             z2, z2_new, r, r_new, gradx_xold, gradx, symgrad_v_vold,
             symgrad_v, tau) = (
             Kyk1_new, Kyk1, Kyk2_new, Kyk2, Ax, Axold, z1_new, z1,
             z2_new, z2, r_new, r, gradx, gradx_xold, symgrad_v,
             symgrad_v_vold, tau_new)

            if not np.mod(i, 50):
                if self.irgn_par["display_iterations"]:
                    self.model.plot_unknowns(x_new.get())
                if self.unknowns_H1 > 0:
                    primal_new = (
                        self.irgn_par["lambd"] / 2 *
                        clarray.vdot(Axold - res, Axold - res) +
                        alpha * clarray.sum(
                            abs((gradx[:self.unknowns_TGV] - v))) +
                        beta * clarray.sum(abs(symgrad_v)) +
                        1 / (2 * delta) * clarray.vdot(
                            x_new - xk, x_new - xk) +
                        self.irgn_par["omega"] / 2 *
                        clarray.vdot(gradx[self.unknowns_TGV:],
                                     gradx[self.unknowns_TGV:])).real

                    dual = (
                        -delta / 2 * clarray.vdot(-Kyk1, -Kyk1)
                        - clarray.vdot(xk, (-Kyk1)) + clarray.sum(Kyk2)
                        - 1 / (2 * self.irgn_par["lambd"]) * clarray.vdot(r, r)
                        - clarray.vdot(res, r)
                        - 1 / (2 * self.irgn_par["omega"])
                        * clarray.vdot(z1[self.unknowns_TGV:],
                                       z1[self.unknowns_TGV:])).real
                else:
                    primal_new = (
                        self.irgn_par["lambd"] / 2 *
                        clarray.vdot(Axold - res, Axold - res) +
                        alpha * clarray.sum(
                            abs((gradx[:self.unknowns_TGV] - v))) +
                        beta * clarray.sum(abs(symgrad_v)) +
                        1 / (2 * delta) * clarray.vdot(
                            x_new - xk, x_new - xk)).real

                    dual = (
                        -delta / 2 * clarray.vdot(-Kyk1, -Kyk1)
                        - clarray.vdot(xk, (-Kyk1)) + clarray.sum(Kyk2)
                        - 1 / (2 * self.irgn_par["lambd"]) * clarray.vdot(r, r)
                        - clarray.vdot(res, r)).real

                gap = np.abs(primal_new - dual)
                if i == 0:
                    gap_init = gap.get()
                if np.abs(primal - primal_new)/self.fval_init <\
                   self.irgn_par["tol"]:
                    print("Terminated at iteration %d because the energy "
                          "decrease in the primal problem was less than %.3e" %
                          (i,
                           np.abs(primal - primal_new).get()/self.fval_init))
                    self.v = v_new.get()
                    self.r = r.get()
                    self.z1 = z1.get()
                    self.z2 = z2.get()
                    return x_new.get()
                if gap > gap_old * self.irgn_par["stag"] and i > 1:
                    self.v = v_new.get()
                    self.r = r.get()
                    self.z1 = z1.get()
                    self.z2 = z2.get()
                    print("Terminated at iteration %d "
                          "because the method stagnated" % (i))
                    return x_new.get()
                if np.abs((gap - gap_old) / gap_init) < self.irgn_par["tol"]:
                    self.v = v_new.get()
                    self.r = r.get()
                    self.z1 = z1.get()
                    self.z2 = z2.get()
                    print("Terminated at iteration %d because the "
                          "relative energy decrease of the PD gap was "
                          "less than %.3e"
                          % (i, np.abs((gap - gap_old).get() / gap_init)))
                    return x_new.get()
                primal = primal_new
                gap_old = gap
                sys.stdout.write(
                    "Iteration: %04d ---- Primal: %2.2e, "
                    "Dual: %2.2e, Gap: %2.2e \r" %
                    (i, 1000*primal.get() / self.fval_init,
                     1000*dual.get() / self.fval_init,
                     1000*gap.get() / self.fval_init))
                sys.stdout.flush()

            (x, x_new) = (x_new, x)
            (v, v_new) = (v_new, v)

        self.v = v.get()
        self.r = r.get()
        self.z1 = z1.get()
        self.z2 = z2.get()
        return x.get()

    def tv_solve_3D(self, x, res, iters):
        alpha = self.irgn_par["gamma"]

        tau = self.tau
        tau_new = np.float32(0)

        x = clarray.to_device(self.queue, x)
        xk = x.copy()
        x_new = clarray.empty_like(x)

        r = clarray.to_device(self.queue, np.zeros_like(self.r))
        r_new = clarray.empty_like(r)
        z1 = clarray.to_device(self.queue, np.zeros_like(self.z1))
        z1_new = clarray.empty_like(z1)
        res = clarray.to_device(self.queue, res.astype(DTYPE))

        delta = self.irgn_par["delta"]
        omega = self.irgn_par["omega"]
        mu = 1 / delta

        theta_line = self.theta_line
        beta_line = self.beta_line
        beta_new = np.float32(0)
        mu_line = np.float32(0.5)
        delta_line = np.float32(1)
        ynorm = np.float32(0.0)
        lhs = np.float32(0.0)
        primal = np.float32(0.0)
        primal_new = np.float32(0)
        dual = np.float32(0.0)
        gap_init = np.float32(0.0)
        gap_old = np.float32(0.0)
        gap = np.float32(0.0)
        self.eval_const()

        Kyk1 = clarray.empty_like(x)
        Kyk1_new = clarray.empty_like(x)
        gradx = clarray.empty_like(z1)
        gradx_xold = clarray.empty_like(z1)
        Axold = clarray.empty_like(res)
        Ax = clarray.empty_like(res)

        Axold.add_event(self.operator_forward(x, Axold))
        Kyk1.add_event(self.operator_adjoint(r, Kyk1, z1))

        for i in range(iters):
            x_new.add_event(self.update_primal(x_new, x, Kyk1, xk, tau, delta))

            beta_new = beta_line * (1 + mu * tau)
            tau_new = tau * np.sqrt(beta_line / beta_new * (1 + theta_line))
            beta_line = beta_new

            gradx.add_event(self.f_grad(
                gradx, x_new, wait_for=gradx.events + x_new.events))
            gradx_xold.add_event(self.f_grad(
                gradx_xold, x, wait_for=gradx_xold.events + x.events))
            Ax.add_event(self.operator_forward(x_new, Ax))

            while True:
                theta_line = tau_new / tau
                z1_new.add_event(self.update_z1_tv(
                    z1_new, z1, gradx, gradx_xold,
                    beta_line * tau_new, theta_line, alpha, omega))
                r_new.add_event(self.update_r(
                    r_new, r, Ax, Axold, res,
                    beta_line * tau_new, theta_line, self.irgn_par["lambd"]))

                Kyk1_new.add_event(self.operator_adjoint(
                    r_new, Kyk1_new, z1_new))

                ynorm = ((clarray.vdot(r_new - r, r_new - r) +
                          clarray.vdot(z1_new - z1, z1_new - z1))**(1 / 2)
                         ).real
                lhs = np.sqrt(beta_line) * tau_new * (
                    (clarray.vdot(Kyk1_new - Kyk1, Kyk1_new - Kyk1))**(1 / 2)
                    ).real
                if lhs <= ynorm * delta_line:
                    break
                else:
                    tau_new = tau_new * mu_line

            (Kyk1, Kyk1_new,  Axold, Ax, z1, z1_new, r, r_new, gradx_xold,
             gradx, tau) = (
             Kyk1_new, Kyk1,  Ax, Axold, z1_new, z1, r_new, r, gradx,
             gradx_xold, tau_new)

            if not np.mod(i, 50):
                if self.irgn_par["display_iterations"]:
                    self.model.plot_unknowns(x_new.get())
                if self.unknowns_H1 > 0:
                    primal_new = (
                        self.irgn_par["lambd"] / 2 *
                        clarray.vdot(Axold - res, Axold - res) +
                        alpha * clarray.sum(
                            abs((gradx[:self.unknowns_TGV]))) +
                        1 / (2 * delta) *
                        clarray.vdot(x_new - xk, x_new - xk) +
                        self.irgn_par["omega"] / 2 *
                        clarray.vdot(gradx[self.unknowns_TGV:],
                                     gradx[self.unknowns_TGV:])).real
                    dual = (
                        -delta / 2 * clarray.vdot(-Kyk1, -Kyk1)
                        - clarray.vdot(xk, (-Kyk1))
                        - 1 / (2 * self.irgn_par["lambd"]) * clarray.vdot(r, r)
                        - clarray.vdot(res, r)
                        - 1 / (2 * self.irgn_par["omega"])
                        * clarray.vdot(z1[self.unknowns_TGV:],
                                       z1[self.unknowns_TGV:])).real
                else:
                    primal_new = (
                        self.irgn_par["lambd"] / 2 *
                        clarray.vdot(Axold - res, Axold - res) +
                        alpha * clarray.sum(
                            abs((gradx[:self.unknowns_TGV]))) +
                        1 / (2 * delta) * clarray.vdot(x_new - xk, x_new - xk)
                        ).real
                    dual = (
                        -delta / 2 * clarray.vdot(-Kyk1, -Kyk1)
                        - clarray.vdot(xk, (-Kyk1))
                        - 1 / (2 * self.irgn_par["lambd"]) * clarray.vdot(r, r)
                        - clarray.vdot(res, r)).real
                gap = np.abs(primal_new - dual)

                if i == 0:
                    gap_init = gap
                if np.abs(primal - primal_new)/self.fval_init < \
                   self.irgn_par["tol"]:
                    print("Terminated at iteration %d because the energy "
                          "decrease in the primal problem was less than %.3e" %
                          (i, np.abs(primal - primal_new).get() /
                              self.fval_init))
                    self.r = r.get()
                    self.z1 = z1.get()
                    return x_new.get()
                if (gap > gap_old * self.irgn_par["stag"]) and i > 1:
                    self.r = r.get()
                    self.z1 = z1.get()
                    print("Terminated at iteration %d "
                          "because the method stagnated" % (i))
                    return x_new.get()
                if np.abs((gap - gap_old) / gap_init) < self.irgn_par["tol"]:
                    self.r = r.get()
                    self.z1 = z1.get()
                    print("Terminated at iteration %d because the relative "
                          "energy decrease of the PD gap was less than %.3e" %
                          (i, np.abs((gap - gap_old).get() / gap_init)))
                    return x_new.get()
                primal = primal_new
                gap_old = gap
                sys.stdout.write(
                    "Iteration: %04d ---- Primal: %2.2e, "
                    "Dual: %2.2e, Gap: %2.2e \r" %
                    (i, 1000 * primal.get() / self.fval_init,
                     1000 * dual.get() / self.fval_init,
                     1000 * gap.get() / self.fval_init))
                sys.stdout.flush()

            (x, x_new) = (x_new, x)

        self.r = r.get()
        self.z1 = z1.get()
        return x.get()

    def execute(self, TV=0, imagespace=0, reco_2D=0):
        if reco_2D:
            print("2D currently not implemented, \
                  3D can be used with a single slice.")
            raise NotImplementedError
        else:
            self.irgn_par["lambd"] *= self.SNR_est
            self.delta = self.irgn_par["delta"]
            self.delta_max = self.irgn_par["delta_max"]
            if imagespace:
                self.execute_3D_imagespace(TV)
            else:
                self.execute_3D(TV)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb 18 08:55:35 2019

@author: omaier

Copyright 2019 Oliver Maier

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from abc import ABC, abstractmethod
import numpy as np


DTYPE = np.complex64
DTYPE_real = np.float32


class constraints:
    def __init__(self, min_val=-np.inf, max_val=np.inf,
                 real_const=False):
        self.min = min_val
        self.max = max_val
        self.real = real_const

    def update(self, scale):
        self.min = self.min / scale
        self.max = self.max / scale


class BaseModel(ABC):
    def __init__(self, par):
        self.constraints = []
        self.uk_scale = []
        self.NScan = par["NScan"]
        self.NSlice = par["NSlice"]
        self.dimX = par["dimX"]
        self.dimY = par["dimY"]
        self.figure = None

    def rescale(self, x):
        tmp_x = np.copy(x)
        for i in range(x.shape[0]):
            tmp_x[i] *= self.uk_scale[i]
        return tmp_x

    def execute_forward(self, x, islice=None):
        if islice is None:
            return self._execute_forward_3D(x)
        else:
            return self._execute_forward_2D(x, islice)

    def execute_gradient(self, x, islice=None):
        if islice is None:
            return self._execute_gradient_3D(x)
        else:
            return self._execute_gradient_2D(x, islice)

    @abstractmethod
    def _execute_forward_2D(self, x, islice):
        ...

    @abstractmethod
    def _execute_gradient_2D(self, x, islice):
        ...

    @abstractmethod
    def _execute_forward_3D(self, x):
        ...

    @abstractmethod
    def _execute_gradient_3D(self, x):
        ...

    @abstractmethod
    def plot_unknowns(self, x, dim_2D=False):
        ...

    @abstractmethod
    def _set_init_scales(self):
        ...
